/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * Chat helper application.
 *
 * @author Christian Grothoff
 * @file applications/chat/gnunet-chat.c 
 **/

#include <unistd.h>
#include "config.h"
#include "applications/common/port.h"
#include "applications/common/printhelp.h"
#include "applications/chat/chat.h"
#include "util/getopt.h"
#include "util/configuration.h"
#include "util/logging.h"
#include "util/initialize.h"
#include "util/returnvalues.h"
#include "util/tcpio.h"
#include "util/semaphore.h"

#define CHAT_VERSION "0.0.1"

static Semaphore * doneSem;

/**
 * Parse the options, set the timeout.
 * @param argc the number of options
 * @param argv the option list (including keywords)
 * @return OK on error, SYSERR if we should exit 
 **/
static int parseOptions(int argc,
			char ** argv) {
  int option_index;
  int c;  
  
  FREENONNULL(setConfigurationString("GNUNETD",
				     "LOGFILE",
				     NULL));
  while (1) {
    static struct option long_options[] = {
      LONG_DEFAULT_OPTIONS,
      { "nickname", 1, 0, 'n' },
      { 0,0,0,0 }
    };    
    option_index=0;
    c = getopt_long(argc,
		    argv, 
		    "vhdc:L:H:n:", 
		    long_options, 
		    &option_index);    
    if (c == -1) 
      break;  // No more flags to process
    if (YES == parseDefaultOptions(c, optarg))
      continue;
    switch(c) {
    case 'n':
      FREENONNULL(setConfigurationString("GNUNET-CHAT",
					 "NICK",
					 optarg));
      break;
    case 'v': 
      printf("GNUnet v%s, gnunet-chat v%s\n",
	     VERSION,
	     CHAT_VERSION);
      return SYSERR;
    case 'h': 
      printf("USAGE: gnunet-chat -n NICKNAME\n");
      printHelpDefault();
      return SYSERR;
    default: 
      LOG(LOG_FAILURE,
	  "FAILURE: Unknown option %c. Aborting.\n"\
	  "Use --help to get a list of options.\n",
	  c);
      return -1;
    } /* end of parsing commandline */
  } /* while (1) */
  return OK;
}

static void receiveThread(GNUNET_TCP_SOCKET * sock) {
  CHAT_CS_MESSAGE * buffer;

  buffer = MALLOC(MAX_BUFFER_SIZE);
  while (OK == readFromSocket(sock, 
			      (CS_HEADER **)&buffer)) {
    if ( (ntohs(buffer->header.size) != sizeof(CHAT_CS_MESSAGE)) ||
	 (ntohs(buffer->header.tcpType) != CHAT_CS_PROTO_MSG) )
      continue;
    buffer->nick[CHAT_NICK_LENGTH-1] = '\0'; 
    buffer->message[CHAT_MSG_LENGTH-1] = '\0';
    printf("%s says: %s",
	   &buffer->nick[0],
	   &buffer->message[0]);
  }
  FREE(buffer);
  SEMAPHORE_UP(doneSem);
}

/**
 * The main function to search for files on GNet.
 * @param argc number of arguments from the command line
 * @param argv command line arguments
 * @return return value from gnunetsearch: 0: ok, -1: error
 **/   
int main(int argc, char ** argv) {
  GNUNET_TCP_SOCKET * sock;
  pthread_t messageReceiveThread;
  CHAT_CS_MESSAGE msg;
  char * nick;

  if (SYSERR == initUtil(argc, argv, &parseOptions))
    return 0; /* parse error, --help, etc. */ 
  sock = getClientSocket();

  nick = getConfigurationString("GNUNET-CHAT", "NICK");
  if (nick == NULL) 
    errexit("You must specify a nickname (use option -n).\n");  

  doneSem = SEMAPHORE_NEW(0);
  PTHREAD_CREATE(&messageReceiveThread, 
		 (void * (*)()) &receiveThread, 
		 sock,
		 128 * 1024);
  pthread_detach(messageReceiveThread);

  memset(&msg, 
	 0, 
	 sizeof(CHAT_CS_MESSAGE));
  memcpy(&msg.message[0], 
	 "Hi!", 
	 strlen("Hi!"));
  msg.header.size 
    = htons(sizeof(CHAT_CS_MESSAGE));
  msg.header.tcpType 
    = htons(CHAT_CS_PROTO_MSG);
  memcpy(&msg.nick[0],
	 nick,
	 strlen(nick));

  /* send first "Hi!" message to gnunetd to indicate "join" */
  if (SYSERR == writeToSocket(sock,
			      &msg.header))
    errexit("FATAL: Could not send join message to gnunetd\n");

  /* read messages from command line and send */
  while (1) {
    memset(&msg.message, 0, 1024);
    if (NULL == fgets(&msg.message[0], 1024, stdin))
      break;
    if (SYSERR == writeToSocket(sock,
				&msg.header))
      errexit("FATAL: Could not send message to gnunetd\n");
  }
  closeSocketTemporarily(sock);
  SEMAPHORE_DOWN(doneSem);
  SEMAPHORE_FREE(doneSem);
  releaseClientSocket(sock);

  doneUtil();
  return 0;
}

/* end of gnunet-chat.c */
