/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/
/**
 * @author Igor Wronsky
 * @author Christian Grothoff
 * @file include/applications/afs/util/bloomfilter.h
 **/
#ifndef BLOOMFILTER_H
#define BLOOMFILTER_H

#include "config.h"
#include "util/hashing.h"
#include "util/semaphore.h"

/* Four bit per entry on 1/4 of the cache size (8MB)
   is 1/4/2 = 1/8 (= 1 MB) */
#define SINGLEBLOOMFILTER_SIZE CACHE_HASH_TABLE_SIZE/8
/* (2/4)^2 is good for now */
#define SINGLEBLOOMFILTER_K    2

/* Eight bits per 24 entries on 3/4 of the cache size
   is 3/4/24 = 1/32 (= 250 kb); since only
   1 out of approx. 25 entries should actually
   be in the filter, this should be enough */
#define SUPERBLOOMFILTER_SIZE  CACHE_HASH_TABLE_SIZE/32
/* (3/8)^3 is good for now */
#define SUPERBLOOMFILTER_K     3

typedef struct {
  /** The bit counter file on disk */
  int fd;
  /** How many bits we set for each stored element */  
  unsigned int addressesPerElement;
  /** The actual bloomfilter bit array */
  char * bitArray;
  /** Size of bitArray in bytes */
  unsigned int bitArraySize;
  /** Concurrency control */
  Mutex lock;
  /** Statistics handle for filter hits */
  int statHandle_hits;
  /** Statistics handle for filter misses */
  int statHandle_misses;
  /** Statistics handle for adds to filter */
  int statHandle_adds;
  /** Statistics handle for dels from filter */
  int statHandle_dels;

} Bloomfilter;

/**
 * Load a bloom-filter from a file.
 * @param filename the name of the file (or the prefix)
 * @param size the size of the bloom-filter (number of
 *        bytes of storage space to use)
 * @param k the number of hash-functions to apply per
 *        element (number of bits set per element in the set)
 * @return the bloomfilter
 **/
Bloomfilter * loadBloomfilter(char * filename,
			      unsigned int size,
			      unsigned int k);

/**
 * Test if an element is in the filter.
 * @param e the element
 * @param bf the filter
 * @return YES if the element is in the filter, NO if not
 **/
int testBloomfilter(Bloomfilter * bf,
		    HashCode160 * e);

/**
 * Add an element to the filter
 * @param bf the filter
 * @param e the element
 **/
void addToBloomfilter(Bloomfilter * bf,
		      HashCode160 * e);

/**
 * Remove an element from the filter.
 * @param bf the filter
 * @param e the element to remove
 **/
void delFromBloomfilter(Bloomfilter * bf,
			HashCode160 * e);

/**
 * Free the space associcated with a filter
 * in memory, flush to drive if needed (do not
 * free the space on the drive)
 * @param bf the filter
 **/
void freeBloomfilter(Bloomfilter * bf);

/**
 * Reset a bloom filter to empty.
 * @param bf the filter
 **/
void resetBloomfilter(Bloomfilter * bf);

#endif 
