/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file include/cron.h
 * @author Christian Grothoff
 *
 * Module to run periodically scheduled or
 * deferred background jobs.
 **/

#ifndef CRON_H
#define CRON_H

#include "config.h"

#include <unistd.h>
#include <pthread.h>

#include "util/semaphore.h"
#include "util/xmalloc.h"

/* use these constants to specify timeings */
#define cronMILLIS 1
#define cronSECONDS 1000 * cronMILLIS
#define cronMINUTES 60 * cronSECONDS
#define cronHOURS 60 * cronMINUTES
#define cronDAYS 24 * cronHOURS
#define cronWEEKS 7 * cronDAYS
#define cronMONTHS 30 * cronDAYS
#define cronYEARS 365 * cronDAYS

/**
 * Time for absolute times used by cron.
 **/
typedef unsigned long long cron_t;  

/**
 * Type of a cron-job method.
 **/
typedef void (*CronJob)(void *);

/**
 * Initialize controlThread.
 **/
void initCron();

/**
 * Make sure to call stopCron before calling this method!
 **/
void doneCron();

/**
 * Start the cron jobs.
 **/
void startCron();

/**
 * Stop the cron service.
 **/
void stopCron();

/**
 * Get the current time (works just as "time", just
 * that we use the unit of time that the cron-jobs use).
 * @param setme will set the current time if non-null
 * @return the current time 
 **/
cron_t cronTime(cron_t * setme);

/**
 * Add a cron-job to the delta list.
 * @param method which method should we run
 * @param delta how many milliseconds until we run the method
 * @param deltaRepeat if this is a periodic, the time between
 *        the runs, otherwise 0.
 * @param data argument to pass to the method
 **/
void addCronJob(CronJob method,
		unsigned int delta,
		unsigned int deltaRepeat,
		void * data);

/**
 * Remove a cron-job from the list. 
 * @param method which method is listed?
 * @param repeat which repeat factor was chosen? 
 * @param data what was the data given to the method
 **/
void delCronJob(CronJob method,
		unsigned int repeat,
		void * data);

/* end of cron.h */
#endif
