/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * Wrapper around malloc/free.
 * @file util/xmalloc.c
 * @author Christian Grothoff
 **/ 

#include "config.h"
#include "util/xmalloc.h"
#include "util/logging.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>


#define DEBUG_MALLOC 0 

/**
 * Allocate memory. Checks the return value, aborts if no more
 * memory is available.
 **/
void * xmalloc(int size, 
	       const char * site) {
  void * result;

  result = malloc(size);
#if DEBUG_MALLOC
  printf("%x malloc %s (%d bytes)\n", 
	 (int)result, 
	 site, 
	 size);
  fflush(stdout);
#endif
  if (result == NULL) {
    errexit("ERROR: out of memory (%d at %s)!\n",
	    size, site);
  }
  memset(result, 0, size); /* client code should not rely on this, though... */
  return result;
}

/**
 * Allocate memory. Checks the return value, aborts if no more
 * memory is available.
 **/
void * xmalloc_(int size, 
		const char * filename,
		const int linenumber) {
  void * result;

  result = malloc(size);
#if DEBUG_MALLOC 
  printf("%x malloc %s:%d (%d bytes)\n", 
	 (int)result, 
	 filename,
	 linenumber,
	 size);
  fflush(stdout);
#endif
  if (result == NULL) {
    errexit("ERROR: out of memory (%d at %s:%d)!\n",
	    size, 
	    filename,
	    linenumber);
  }
  memset(result, 0, size); /* client code should not rely on this, though... */
  return result;
}

/**
 * Free memory. Merely a wrapper for the case that we
 * want to keep track of allocations.
 **/
void xfree(void * ptr, 
	   const char * site) {
  if (ptr == NULL)
    errexit("xfree called with NULL ptr at %s\n", 
	    site);
#if DEBUG_MALLOC
  printf("%x free %s\n", (int) ptr, site);
  fflush(stdout);
#endif
  free(ptr);
}

/**
 * Free memory. Merely a wrapper for the case that we
 * want to keep track of allocations.
 **/
void xfree_(void * ptr, 
	    const char * filename,
	    const int linenumber) {
  if (ptr == NULL)
    errexit("FREE called with NULL ptr at %s:%d\n", 
	    filename,
	    linenumber);
#if DEBUG_MALLOC 
  printf("%x free %s:%d\n", 
	 (int) ptr, 
	 filename,
	 linenumber);
  fflush(stdout);
#endif
  free(ptr);
}

/**
 * Free if not NULL.
 **/
void freeNonNull(void * ptr,
		 const char * site) {
  if (ptr != NULL)
    xfree(ptr, site);
}

/**
 * Dup a string.
 **/
char * xstrdup_(const char * str,
		const char * filename,
		const int linenumber) {
  char * res;

  res = strdup(str);
#if DEBUG_MALLOC 
  printf("%x strdup %s:%d (%d bytes)\n", 
	 (int) res, 
	 filename,
	 linenumber,
	 strlen(str)+1);
  fflush(stdout);
#endif
  return res;
}

/**
 * Grow an array.
 * Grows old by (*oldCount-newCount)*elementSize
 * bytes and sets *oldCount to newCount.
 * @param old address of the pointer to the array
 *        *old may be NULL
 * @param elementSize the size of the elements of the array
 * @param oldCount address of the number of elements in the *old array
 * @param newCount number of elements in the new array, may be 0
 **/
void xgrow_(void ** old,
	    int elementSize,
	    int * oldCount,
	    int newCount,
	    char * filename,
	    const int linenumber) {
  void * tmp;
  int size;

  size = newCount * elementSize;
  if (size == 0)
    tmp = NULL;
  else {
    tmp = malloc(size);  
#if DEBUG_MALLOC 
    printf("%x malloc %s:%d\n", 
	   (int)tmp, 
	   filename,
	   linenumber);
    fflush(stdout);
#endif
    if (tmp == NULL) {
      errexit("ERROR: out of memory (%d at %s:%d)!\n",
	      size, 
	      filename,
	      linenumber);
    }
    memset(tmp, 0, size); /* client code should not rely on this, though... */
    if (*oldCount > newCount)
      *oldCount = newCount; /* shrink is also allowed! */
    memcpy(tmp, 
	   *old, 
	   elementSize * (*oldCount));
  }

  if (*old != NULL) {
#if DEBUG_MALLOC 
    printf("%x free %s:%d\n", 
	   (int) *old, 
	   filename,
	   linenumber);
    fflush(stdout);
#endif
    free(*old);
  }
  *old = tmp;
  *oldCount = newCount;
}

/* end of xmalloc.c */
